@@ -17,6 +17,7 @@ source ../support/test.tcl
 
 set ::verbose 0
 set ::valgrind 0
+set ::tls 0
 set ::pause_on_error 0
 set ::simulate_error 0
 set ::failed 0
@@ -69,7 +70,19 @@ proc spawn_instance {type base_port count {conf {}}} {
         # Write the instance config file.
         set cfgfile [file join $dirname $type.conf]
         set cfg [open $cfgfile w]
-        puts $cfg "port $port"
+        if {$::tls} {
+            puts $cfg "tls-port $port"
+            puts $cfg "tls-replication yes"
+            puts $cfg "tls-cluster yes"
+            puts $cfg "port 0"
+            puts $cfg [format "tls-cert-file %s/../../tls/redis.crt" [pwd]]
+            puts $cfg [format "tls-key-file %s/../../tls/redis.key" [pwd]]
+            puts $cfg [format "tls-dh-params-file %s/../../tls/redis.dh" [pwd]]
+            puts $cfg [format "tls-ca-cert-file %s/../../tls/ca.crt" [pwd]]
+            puts $cfg "loglevel debug"
+        } else {
+            puts $cfg "port $port"
+        }
         puts $cfg "dir ./$dirname"
         puts $cfg "logfile log.txt"
         # Add additional config files
@@ -84,13 +97,11 @@ proc spawn_instance {type base_port count {conf {}}} {
 
         # Check availability
         if {[server_is_up 127.0.0.1 $port 100] == 0} {
-            set logfile [file join $dirname log.txt]
-            puts [exec tail $logfile]
-            abort_sentinel_test "Problems starting $type #$j: ping timeout, maybe server start failed, check $logfile"
+            abort_sentinel_test "Problems starting $type #$j: ping timeout"
         }
 
         # Push the instance into the right list
-        set link [redis 127.0.0.1 $port]
+        set link [redis 127.0.0.1 $port 0 $::tls]
         $link reconnect 1
         lappend ::${type}_instances [list \
             pid $pid \
@@ -150,6 +161,13 @@ proc parse_options {} {
             set ::simulate_error 1
         } elseif {$opt eq {--valgrind}} {
             set ::valgrind 1
+        } elseif {$opt eq {--tls}} {
+            package require tls 1.6
+            ::tls::init \
+                -cafile "$::tlsdir/ca.crt" \
+                -certfile "$::tlsdir/redis.crt" \
+                -keyfile "$::tlsdir/redis.key"
+            set ::tls 1
         } elseif {$opt eq "--help"} {
             puts "Hello, I'm sentinel.tcl and I run Sentinel unit tests."
             puts "\nOptions:"
@@ -336,16 +354,10 @@ proc S {n args} {
     [dict get $s link] {*}$args
 }
 
-# Returns a Redis instance by index.
-# Example:
-#     [Rn 0] info
-proc Rn {n} {
-    return [dict get [lindex $::redis_instances $n] link]
-}
-
 # Like R but to chat with Redis instances.
 proc R {n args} {
-    [Rn $n] {*}$args
+    set r [lindex $::redis_instances $n]
+    [dict get $r link] {*}$args
 }
 
 proc get_info_field {info field} {
@@ -464,12 +476,12 @@ proc kill_instance {type id} {
 
     # Wait for the port it was using to be available again, so that's not
     # an issue to start a new server ASAP with the same port.
-    set retry 100
+    set retry 10
     while {[incr retry -1]} {
-        set port_is_free [catch {set s [socket 127.0.0.1 $port]}]
+        set port_is_free [catch {set s [socket 127.0.01 $port]}]
         if {$port_is_free} break
         catch {close $s}
-        after 100
+        after 1000
     }
     if {$retry == 0} {
         error "Port $port does not return available after killing instance."
@@ -496,13 +508,11 @@ proc restart_instance {type id} {
 
     # Check that the instance is running
     if {[server_is_up 127.0.0.1 $port 100] == 0} {
-        set logfile [file join $dirname log.txt]
-        puts [exec tail $logfile]
-        abort_sentinel_test "Problems starting $type #$id: ping timeout, maybe server start failed, check $logfile"
+        abort_sentinel_test "Problems starting $type #$id: ping timeout"
     }
 
     # Connect with it with a fresh link
-    set link [redis 127.0.0.1 $port]
+    set link [redis 127.0.0.1 $port 0 $::tls]
     $link reconnect 1
     set_instance_attrib $type $id link $link
 
@@ -519,16 +529,3 @@ proc restart_instance {type id} {
     }
 }
 
-proc redis_deferring_client {type id} {
-    set port [get_instance_attrib $type $id port]
-    set host [get_instance_attrib $type $id host]
-    set client [redis $host $port 1]
-    return $client
-}
-
-proc redis_client {type id} {
-    set port [get_instance_attrib $type $id port]
-    set host [get_instance_attrib $type $id host]
-    set client [redis $host $port 0]
-    return $client
-}
