@@ -219,4 +219,17 @@ start_server {tags {"expire"}} {
         set ttl [r ttl foo]
         assert {$ttl <= 98 && $ttl > 90}
     }
+
+    test {SET command will remove expire} {
+        r set foo bar EX 100
+        r set foo bar
+        r ttl foo
+    } {-1}
+
+    test {SET - use KEEPTTL option, TTL should not be removed} {
+        r set foo bar EX 100
+        r set foo bar KEEPTTL
+        set ttl [r ttl foo]
+        assert {$ttl <= 100 && $ttl > 90}
+    }
 }
